# type: ignore
# TODO: 1. remove the line above after implementing this file
# TODO: 2. remove these comments

from __future__ import annotations

from qadence2_ir.types import Model

from .interface import Interface


def compile_to_backend(model: Model) -> Interface:
    # TODO: 1. implement this method logic
    # TODO: 2. do not add more arguments
    # TODO: 3. write a docstring if needed
    # TODO: 4. fill the return below with the respective arguments
    # TODO: 5. erase these TODO comments
    return Interface()
